#!/usr/bin/env python3
from gi.repository import Gtk, GObject
import asyncio
import gbulb

class ProgressBarWindow(Gtk.Window):

    def __init__(self):
        Gtk.Window.__init__(self, title="ProgressBar Demo")
        self.set_border_width(10)

        vbox = Gtk.VBox()
        self.add(vbox)

        self.progressbar = Gtk.ProgressBar()
        vbox.pack_start(self.progressbar, True, True, 0)

        button = Gtk.Button("Magic button")
        button.connect("clicked", self.on_magic)
        vbox.pack_start(button, True, True, 0)
        self._magic_button = button

        button = Gtk.Button("Stop button")
        button.connect("clicked", self.on_stop)
        vbox.pack_start(button, True, True, 0)
        self._stop_button = button

        self._running = False

    def on_show_text_toggled(self, button):
        self.progressbar.set_text(text)
        self.progressbar.set_show_text(show_text)

    def on_activity_mode_toggled(self, button):
        self.progressbar.set_fraction(0.0)

    def on_right_to_left_toggled(self, button):
        value = button.get_active()
        self.progressbar.set_inverted(value)

    def on_magic(self, button):
        def coro():
            try:
                yield from gbulb.wait_signal(self._magic_button, "clicked")
                self.progressbar.set_text ("blah blah!")
                self.progressbar.set_fraction(0.50)

                r = yield from asyncio.sleep(1)

                self.progressbar.set_fraction(0.75)
                self.progressbar.set_text ("pouet pouet!")

                r = yield from gbulb.wait_signal(self._magic_button, "clicked")

                self.progressbar.set_fraction(1.0)
                self.progressbar.set_text ("done!")

                yield from asyncio.sleep(1)

            finally:
                self.progressbar.set_fraction(0.0)
                self.progressbar.set_show_text(False)
                self._running = False

        if not self._running:
            self.progressbar.set_fraction(0.25)
            self.progressbar.set_text ("do some magic!")
            self.progressbar.set_show_text (True)
            self._running = asyncio.async(coro())
    
    def on_stop(self, button):
        if self._running:
            self._running.cancel()

            
asyncio.set_event_loop_policy(gbulb.GtkEventLoopPolicy())

win = ProgressBarWindow()
win.connect("delete-event", Gtk.main_quit)
win.show_all()

asyncio.get_event_loop().run_forever()





"""Base implementation of event loop.

The event loop can be broken up into a multiplexer (the part
responsible for notifying us of IO events) and the event loop proper,
which wraps a multiplexer with functionality for scheduling callbacks,
immediately or at a given time in the future.

Whenever a public API takes a callback, subsequent positional
arguments will be passed to the callback if/when it is called.  This
avoids the proliferation of trivial lambdas implementing closures.
Keyword arguments for the callback are not supported; this is a
conscious design decision, leaving the door open for keyword arguments
to modify the meaning of the API call itself.
"""


import collections
import concurrent.futures
#import heapq
import logging
import socket
import subprocess
import time
import os
import sys

from asyncio import events
from asyncio import futures
from asyncio import tasks
from asyncio.log import logger
from asyncio.base_events import Server


__all__ = ['BaseEventLoop']


# Argument for default thread pool executor creation.
_MAX_WORKERS = 5


#class _StopError(BaseException):
#    """Raised to stop the event loop."""
#
#
#def _raise_stop_error(*args):
#    raise _StopError


#class Server(events.AbstractServer):
#
#    def __init__(self, loop, sockets):
#        self.loop = loop
#        self.sockets = sockets
#        self.active_count = 0
#        self.waiters = []
#
#    def attach(self, transport):
#        assert self.sockets is not None
#        self.active_count += 1
#
#    def detach(self, transport):
#        assert self.active_count > 0
#        self.active_count -= 1
#        if self.active_count == 0 and self.sockets is None:
#            self._wakeup()
#
#    def close(self):
#        sockets = self.sockets
#        if sockets is not None:
#            self.sockets = None
#            for sock in sockets:
#                self.loop._stop_serving(sock)
#            if self.active_count == 0:
#                self._wakeup()
#
#    def _wakeup(self):
#        waiters = self.waiters
#        self.waiters = None
#        for waiter in waiters:
#            if not waiter.done():
#                waiter.set_result(waiter)
#
#    @tasks.coroutine
#    def wait_closed(self):
#        if self.sockets is None or self.waiters is None:
#            return
#        waiter = futures.Future(loop=self.loop)
#        self.waiters.append(waiter)
#        yield from waiter


class BaseEventLoop(events.AbstractEventLoop):

    def __init__(self):
#        self._ready = collections.deque()
#        self._scheduled = []
        self._default_executor = None
        self._internal_fds = 0
#        self._running = False

    def _make_socket_transport(self, sock, protocol, waiter=None, *,
                               extra=None, server=None):
        """Create socket transport."""
        raise NotImplementedError

    def _make_ssl_transport(self, rawsock, protocol, sslcontext, waiter, *,
                            server_side=False, server_hostname=None,
                            extra=None, server=None):
        """Create SSL transport."""
        raise NotImplementedError

    def _make_datagram_transport(self, sock, protocol,
                                 address=None, extra=None):
        """Create datagram transport."""
        raise NotImplementedError

    def _make_read_pipe_transport(self, pipe, protocol, waiter=None,
                                  extra=None):
        """Create read pipe transport."""
        raise NotImplementedError

    def _make_write_pipe_transport(self, pipe, protocol, waiter=None,
                                   extra=None):
        """Create write pipe transport."""
        raise NotImplementedError

    @tasks.coroutine
    def _make_subprocess_transport(self, protocol, args, shell,
                                   stdin, stdout, stderr, bufsize,
                                   extra=None, **kwargs):
        """Create subprocess transport."""
        raise NotImplementedError

    def _read_from_self(self):
        """XXX"""
        raise NotImplementedError

    def _write_to_self(self):
        """XXX"""
        raise NotImplementedError

#    def _process_events(self, event_list):
#        """Process selector events."""
#        raise NotImplementedError
#
#    def run_forever(self):
#        """Run until stop() is called."""
#        if self._running:
#            raise RuntimeError('Event loop is running.')
#        self._running = True
#        try:
#            while True:
#                try:
#                    self._run_once()
#                except _StopError:
#                    break
#        finally:
#            self._running = False
#
#    def run_until_complete(self, future):
#        """Run until the Future is done.
#
#        If the argument is a coroutine, it is wrapped in a Task.
#
#        XXX TBD: It would be disastrous to call run_until_complete()
#        with the same coroutine twice -- it would wrap it in two
#        different Tasks and that can't be good.
#
#        Return the Future's result, or raise its exception.
#        """
#        future = tasks.async(future, loop=self)
#        future.add_done_callback(_raise_stop_error)
#        self.run_forever()
#        future.remove_done_callback(_raise_stop_error)
#        if not future.done():
#            raise RuntimeError('Event loop stopped before Future completed.')
#
#        return future.result()
#
#    def stop(self):
#        """Stop running the event loop.
#
#        Every callback scheduled before stop() is called will run.
#        Callback scheduled after stop() is called won't.  However,
#        those callbacks will run if run() is called again later.
#        """
#        self.call_soon(_raise_stop_error)

    def close(self):
        """Close the event loop.

        This clears the queues and shuts down the executor,
        but does not wait for the executor to finish.
        """
#        self._ready.clear()
#        self._scheduled.clear()
        executor = self._default_executor
        if executor is not None:
            self._default_executor = None
            executor.shutdown(wait=False)

#    def is_running(self):
#        """Returns running status of event loop."""
#        return self._running
#
#    def time(self):
#        """Return the time according to the event loop's clock."""
#        return time.monotonic()
#
#    def call_later(self, delay, callback, *args):
#        """Arrange for a callback to be called at a given time.
#
#        Return a Handle: an opaque object with a cancel() method that
#        can be used to cancel the call.
#
#        The delay can be an int or float, expressed in seconds.  It is
#        always a relative time.
#
#        Each callback will be called exactly once.  If two callbacks
#        are scheduled for exactly the same time, it undefined which
#        will be called first.
#
#        Any positional arguments after the callback will be passed to
#        the callback when it is called.
#        """
#        return self.call_at(self.time() + delay, callback, *args)
#
#    def call_at(self, when, callback, *args):
#        """Like call_later(), but uses an absolute time."""
#        timer = events.TimerHandle(when, callback, args)
#        heapq.heappush(self._scheduled, timer)
#        return timer
#
#    def call_soon(self, callback, *args):
#        """Arrange for a callback to be called as soon as possible.
#
#        This operates as a FIFO queue, callbacks are called in the
#        order in which they are registered.  Each callback will be
#        called exactly once.
#
#        Any positional arguments after the callback will be passed to
#        the callback when it is called.
#        """
#        handle = events.Handle(callback, args)
#        self._ready.append(handle)
#        return handle

    def call_soon_threadsafe(self, callback, *args):
        """XXX"""
        handle = self.call_soon(callback, *args)
        self._write_to_self()
        return handle

    def run_in_executor(self, executor, callback, *args):
        if isinstance(callback, events.Handle):
            assert not args
            assert not isinstance(callback, events.TimerHandle)
            if callback._cancelled:
                f = futures.Future(loop=self)
                f.set_result(None)
                return f
            callback, args = callback._callback, callback._args
        if executor is None:
            executor = self._default_executor
            if executor is None:
                executor = concurrent.futures.ThreadPoolExecutor(_MAX_WORKERS)
                self._default_executor = executor
        return futures.wrap_future(executor.submit(callback, *args), loop=self)

    def set_default_executor(self, executor):
        self._default_executor = executor

    def getaddrinfo(self, host, port, *,
                    family=0, type=0, proto=0, flags=0):
        return self.run_in_executor(None, socket.getaddrinfo,
                                    host, port, family, type, proto, flags)

    def getnameinfo(self, sockaddr, flags=0):
        return self.run_in_executor(None, socket.getnameinfo, sockaddr, flags)

    @tasks.coroutine
    def create_connection(self, protocol_factory, host=None, port=None, *,
                          ssl=None, family=0, proto=0, flags=0, sock=None,
                          local_addr=None, server_hostname=None):
        """XXX"""
        if server_hostname is not None and not ssl:
            raise ValueError('server_hostname is only meaningful with ssl')

        if server_hostname is None and ssl:
            # Use host as default for server_hostname.  It is an error
            # if host is empty or not set, e.g. when an
            # already-connected socket was passed or when only a port
            # is given.  To avoid this error, you can pass
            # server_hostname='' -- this will bypass the hostname
            # check.  (This also means that if host is a numeric
            # IP/IPv6 address, we will attempt to verify that exact
            # address; this will probably fail, but it is possible to
            # create a certificate for a specific IP address, so we
            # don't judge it here.)
            if not host:
                raise ValueError('You must set server_hostname '
                                 'when using ssl without a host')
            server_hostname = host

        if host is not None or port is not None:
            if sock is not None:
                raise ValueError(
                    'host/port and sock can not be specified at the same time')

            f1 = self.getaddrinfo(
                host, port, family=family,
                type=socket.SOCK_STREAM, proto=proto, flags=flags)
            fs = [f1]
            if local_addr is not None:
                f2 = self.getaddrinfo(
                    *local_addr, family=family,
                    type=socket.SOCK_STREAM, proto=proto, flags=flags)
                fs.append(f2)
            else:
                f2 = None

            yield from tasks.wait(fs, loop=self)

            infos = f1.result()
            if not infos:
                raise OSError('getaddrinfo() returned empty list')
            if f2 is not None:
                laddr_infos = f2.result()
                if not laddr_infos:
                    raise OSError('getaddrinfo() returned empty list')

            exceptions = []
            for family, type, proto, cname, address in infos:
                try:
                    sock = socket.socket(family=family, type=type, proto=proto)
                    sock.setblocking(False)
                    if f2 is not None:
                        for _, _, _, _, laddr in laddr_infos:
                            try:
                                sock.bind(laddr)
                                break
                            except OSError as exc:
                                exc = OSError(
                                    exc.errno, 'error while '
                                    'attempting to bind on address '
                                    '{!r}: {}'.format(
                                        laddr, exc.strerror.lower()))
                                exceptions.append(exc)
                        else:
                            sock.close()
                            sock = None
                            continue
                    yield from self.sock_connect(sock, address)
                except OSError as exc:
                    if sock is not None:
                        sock.close()
                    exceptions.append(exc)
                else:
                    break
            else:
                if len(exceptions) == 1:
                    raise exceptions[0]
                else:
                    # If they all have the same str(), raise one.
                    model = str(exceptions[0])
                    if all(str(exc) == model for exc in exceptions):
                        raise exceptions[0]
                    # Raise a combined exception so the user can see all
                    # the various error messages.
                    raise OSError('Multiple exceptions: {}'.format(
                        ', '.join(str(exc) for exc in exceptions)))

        elif sock is None:
            raise ValueError(
                'host and port was not specified and no sock specified')

        sock.setblocking(False)

        protocol = protocol_factory()
        waiter = futures.Future(loop=self)
        if ssl:
            sslcontext = None if isinstance(ssl, bool) else ssl
            transport = self._make_ssl_transport(
                sock, protocol, sslcontext, waiter,
                server_side=False, server_hostname=server_hostname)
        else:
            transport = self._make_socket_transport(sock, protocol, waiter)

        yield from waiter
        return transport, protocol

    @tasks.coroutine
    def create_datagram_endpoint(self, protocol_factory,
                                 local_addr=None, remote_addr=None, *,
                                 family=0, proto=0, flags=0):
        """Create datagram connection."""
        if not (local_addr or remote_addr):
            if family == 0:
                raise ValueError('unexpected address family')
            addr_pairs_info = (((family, proto), (None, None)),)
        else:
            # join addresss by (family, protocol)
            addr_infos = collections.OrderedDict()
            for idx, addr in ((0, local_addr), (1, remote_addr)):
                if addr is not None:
                    assert isinstance(addr, tuple) and len(addr) == 2, (
                        '2-tuple is expected')

                    infos = yield from self.getaddrinfo(
                        *addr, family=family, type=socket.SOCK_DGRAM,
                        proto=proto, flags=flags)
                    if not infos:
                        raise OSError('getaddrinfo() returned empty list')

                    for fam, _, pro, _, address in infos:
                        key = (fam, pro)
                        if key not in addr_infos:
                            addr_infos[key] = [None, None]
                        addr_infos[key][idx] = address

            # each addr has to have info for each (family, proto) pair
            addr_pairs_info = [
                (key, addr_pair) for key, addr_pair in addr_infos.items()
                if not ((local_addr and addr_pair[0] is None) or
                        (remote_addr and addr_pair[1] is None))]

            if not addr_pairs_info:
                raise ValueError('can not get address information')

        exceptions = []

        for ((family, proto),
             (local_address, remote_address)) in addr_pairs_info:
            sock = None
            r_addr = None
            try:
                sock = socket.socket(
                    family=family, type=socket.SOCK_DGRAM, proto=proto)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.setblocking(False)

                if local_addr:
                    sock.bind(local_address)
                if remote_addr:
                    yield from self.sock_connect(sock, remote_address)
                    r_addr = remote_address
            except OSError as exc:
                if sock is not None:
                    sock.close()
                exceptions.append(exc)
            else:
                break
        else:
            raise exceptions[0]

        protocol = protocol_factory()
        transport = self._make_datagram_transport(sock, protocol, r_addr)
        return transport, protocol

    @tasks.coroutine
    def create_server(self, protocol_factory, host=None, port=None,
                      *,
                      family=socket.AF_UNSPEC,
                      flags=socket.AI_PASSIVE,
                      sock=None,
                      backlog=100,
                      ssl=None,
                      reuse_address=None):
        """XXX"""
        if isinstance(ssl, bool):
            raise TypeError('ssl argument must be an SSLContext or None')
        if host is not None or port is not None:
            if sock is not None:
                raise ValueError(
                    'host/port and sock can not be specified at the same time')

            AF_INET6 = getattr(socket, 'AF_INET6', 0)
            if reuse_address is None:
                reuse_address = os.name == 'posix' and sys.platform != 'cygwin'
            sockets = []
            if host == '':
                host = None

            infos = yield from self.getaddrinfo(
                host, port, family=family,
                type=socket.SOCK_STREAM, proto=0, flags=flags)
            if not infos:
                raise OSError('getaddrinfo() returned empty list')

            completed = False
            try:
                for res in infos:
                    af, socktype, proto, canonname, sa = res
                    try:
                        sock = socket.socket(af, socktype, proto)
                    except socket.error:
                        # Assume it's a bad family/type/protocol combination.
                        continue
                    sockets.append(sock)
                    if reuse_address:
                        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR,
                                        True)
                    # Disable IPv4/IPv6 dual stack support (enabled by
                    # default on Linux) which makes a single socket
                    # listen on both address families.
                    if af == AF_INET6 and hasattr(socket, 'IPPROTO_IPV6'):
                        sock.setsockopt(socket.IPPROTO_IPV6,
                                        socket.IPV6_V6ONLY,
                                        True)
                    try:
                        sock.bind(sa)
                    except OSError as err:
                        raise OSError(err.errno, 'error while attempting '
                                      'to bind on address %r: %s'
                                      % (sa, err.strerror.lower()))
                completed = True
            finally:
                if not completed:
                    for sock in sockets:
                        sock.close()
        else:
            if sock is None:
                raise ValueError(
                    'host and port was not specified and no sock specified')
            sockets = [sock]

        server = Server(self, sockets)
        for sock in sockets:
            sock.listen(backlog)
            sock.setblocking(False)
            self._start_serving(protocol_factory, sock, ssl, server)
        return server

    @tasks.coroutine
    def connect_read_pipe(self, protocol_factory, pipe):
        protocol = protocol_factory()
        waiter = futures.Future(loop=self)
        transport = self._make_read_pipe_transport(pipe, protocol, waiter)
        yield from waiter
        return transport, protocol

    @tasks.coroutine
    def connect_write_pipe(self, protocol_factory, pipe):
        protocol = protocol_factory()
        waiter = futures.Future(loop=self)
        transport = self._make_write_pipe_transport(pipe, protocol, waiter)
        yield from waiter
        return transport, protocol

    @tasks.coroutine
    def subprocess_shell(self, protocol_factory, cmd, *, stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         universal_newlines=False, shell=True, bufsize=0,
                         **kwargs):
        if not isinstance(cmd, str):
            raise ValueError("cmd must be a string")
        if universal_newlines:
            raise ValueError("universal_newlines must be False")
        if not shell:
            raise ValueError("shell must be True")
        if bufsize != 0:
            raise ValueError("bufsize must be 0")
        protocol = protocol_factory()
        transport = yield from self._make_subprocess_transport(
            protocol, cmd, True, stdin, stdout, stderr, bufsize, **kwargs)
        return transport, protocol

    @tasks.coroutine
    def subprocess_exec(self, protocol_factory, *args, stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                        universal_newlines=False, shell=False, bufsize=0,
                        **kwargs):
        if universal_newlines:
            raise ValueError("universal_newlines must be False")
        if shell:
            raise ValueError("shell must be False")
        if bufsize != 0:
            raise ValueError("bufsize must be 0")
        protocol = protocol_factory()
        transport = yield from self._make_subprocess_transport(
            protocol, args, False, stdin, stdout, stderr, bufsize, **kwargs)
        return transport, protocol

#    def _add_callback(self, handle):
#        """Add a Handle to ready or scheduled."""
#        assert isinstance(handle, events.Handle), 'A Handle is required here'
#        if handle._cancelled:
#            return
#        if isinstance(handle, events.TimerHandle):
#            heapq.heappush(self._scheduled, handle)
#        else:
#            self._ready.append(handle)
#
#    def _add_callback_signalsafe(self, handle):
#        """Like _add_callback() but called from a signal handler."""
#        self._add_callback(handle)
#        self._write_to_self()
#
#    def _run_once(self):
#        """Run one full iteration of the event loop.
#
#        This calls all currently ready callbacks, polls for I/O,
#        schedules the resulting callbacks, and finally schedules
#        'call_later' callbacks.
#        """
#        # Remove delayed calls that were cancelled from head of queue.
#        while self._scheduled and self._scheduled[0]._cancelled:
#            heapq.heappop(self._scheduled)
#
#        timeout = None
#        if self._ready:
#            timeout = 0
#        elif self._scheduled:
#            # Compute the desired timeout.
#            when = self._scheduled[0]._when
#            deadline = max(0, when - self.time())
#            if timeout is None:
#                timeout = deadline
#            else:
#                timeout = min(timeout, deadline)
#
#        # TODO: Instrumentation only in debug mode?
#        if logger.isEnabledFor(logging.INFO):
#            t0 = self.time()
#            event_list = self._selector.select(timeout)
#            t1 = self.time()
#            if t1-t0 >= 1:
#                level = logging.INFO
#            else:
#                level = logging.DEBUG
#            if timeout is not None:
#                logger.log(level, 'poll %.3f took %.3f seconds',
#                           timeout, t1-t0)
#            else:
#                logger.log(level, 'poll took %.3f seconds', t1-t0)
#        else:
#            t0 = self.time()
#            event_list = self._selector.select(timeout)
#            dt = self.time() - t0
#            if not event_list and timeout and dt < timeout:
#                print("asyncio: selector.select(%.3f ms) took %.3f ms"
#                      % (timeout*1e3, dt*1e3),
#                      file=sys.__stderr__, flush=True)
#        self._process_events(event_list)
#
#        # Handle 'later' callbacks that are ready.
#        now = self.time()
#        while self._scheduled:
#            handle = self._scheduled[0]
#            if handle._when > now:
#                break
#            handle = heapq.heappop(self._scheduled)
#            self._ready.append(handle)
#
#        # This is the only place where callbacks are actually *called*.
#        # All other places just add them to ready.
#        # Note: We run all currently scheduled callbacks, but not any
#        # callbacks scheduled by callbacks run this time around --
#        # they will be run the next time (after another I/O poll).
#        # Use an idiom that is threadsafe without using locks.
#        ntodo = len(self._ready)
#        for i in range(ntodo):
#            handle = self._ready.popleft()
#            if not handle._cancelled:
#                handle._run()
#        handle = None  # Needed to break cycles when an exception occurs.



# vim:sw=4:sts=4:nosta:et:
"""PEP 3156 event loop based on GLib"""

from gi.repository import GLib, GObject, Gio
try:
    from gi.repository import Gtk
except ImportError:
    Gtk = None

from asyncio import events
from asyncio import futures
from asyncio import tasks
from asyncio.log import logger

from . import unix_events

import threading
import signal
import weakref
import collections
import os

class GLibChildWatcher(unix_events.AbstractChildWatcher):
    def __init__(self):
        self._sources = {}

    def attach_loop(self, loop):
        # just ignored
        pass

    def add_child_handler(self, pid, callback, *args):
        self.remove_child_handler(pid)

        source = GLib.child_watch_add(0, pid, self._glib_callback)
        self._sources[pid] = source, callback, args

    def remove_child_handler(self, pid):
        try:
            source = self._sources.pop(pid)[0]
        except KeyError:
            return False

        GLib.source_remove(source)
        return True

    def close(self):
        for source, callback, args in self._sources.values():
            GLib.source_remove(source)

    def __enter__(self):
        return self

    def __exit__(self, a, b, c):
        pass

    def _glib_callback(self, pid, status):

        try:
            source, callback, args = self._sources.pop(pid)
        except KeyError:
            return

        GLib.source_remove(source)

        if os.WIFSIGNALED(status):
            returncode = -os.WTERMSIG(status)
        elif os.WIFEXITED(status):
            returncode = os.WEXITSTATUS(status)

            #FIXME: Hack for adjusting invalid status returned by GLIB
            #	Looks like there is a bug in glib or in pygobject
            if returncode > 128:
                returncode = 128 - returncode
        else:
            returncode = status

        callback(pid, returncode, *args)


class GLibHandle(events.Handle):
    def __init__(self, loop, source, repeat, callback, args):
        super().__init__(callback, args, loop)

        self._loop   = loop
        self._source = source
        self._repeat = repeat
        self._ready  = False
        source.set_callback(self.__class__._callback, self)
        source.attach(loop._context)
        loop._handlers.add(self)

    def cancel(self):
        super().cancel()
        self._source.destroy()
        self._loop._handlers.discard(self)

    def _run(self):
        self._ready = False
        super()._run()

    def _callback(self):
        if not self._ready:
            self._ready = True
            self._loop._ready.append(self)

        self._loop._dispatch()

        if not self._repeat:
            self._loop._handlers.discard(self)
        return self._repeat

#
# Divergences with PEP 3156
#
# In GLib, the concept of event loop is split in two classes: GLib.MainContext
# and GLib.MainLoop.
#
# The thing is mostly implemented by MainContext. MainLoop is just a wrapper
# that implements the run() and quit() functions. MainLoop.run() atomically
# acquires a MainContext and repeatedly calls MainContext.iteration() until
# MainLoop.quit() is called.
#
# A MainContext is not bound to a particular thread, however is cannot be used
# by multiple threads concurrently. If the context is owned by another thread,
# then MainLoop.run() will block until the context is released by the other
# thread.
#
# MainLoop.run() may be called recursively by the same thread (this is mainly
# used for implementing modal dialogs in Gtk).
#
#
# The issue: given a context, GLib provides no ways to know if there is an
# existing event loop running for that context. It implies the following
# divergences with PEP 3156:
#
#  - .run_forever() and .run_until_complete() are not guaranteed to run
#    immediatly. If the context is owned by another thread, then they will
#    block until the context is released by the other thread.
#
#  - .stop() is relevant only when the currently running Glib.MainLoop object
#    was created by this asyncio object (i.e. by calling .run_forever() or
#    .run_until_complete()). The event loop will quit only when it regains
#    control of the context. This can happen in two cases:
#     1. when multiple event loop are enclosed (by creating new MainLoop
#        objects and calling .run() recursively)
#     2. when the event loop has not even yet started because it is still
#        trying to acquire the context
#
# It should be wiser not to use any recursion at all. GLibEventLoop will
# actually prevent you from doing that (in accordance with PEP 3156). However
# you should keep in mind that enclosed loops may be started at any time by
# third-party code calling directly GLib's primitives.
#
#
# TODO: documentation about signal GLib allows catching signals from any
# thread. It is dispatched to the first handler whose flag is not yet raised.
#
# about SIGINT -> KeyboardInterrupt will never be raised asynchronously

class BaseGLibEventLoop(unix_events.SelectorEventLoop):
    """GLib base event loop

    This class handles only the operations related to Glib.MainContext objects.

    Glib.MainLoop operations are implemented in the derived classes.
    """

    class DefaultSigINTHandler:
        def __init__(self):
            s = GLib.unix_signal_source_new(signal.SIGINT)
            s.set_callback(self.__class__._callback, self)
            s.attach()

            self._source = s
            self._loop   = None

        def attach(self, loop):
            if self._loop:
                l = self._loop()
                if l and l != loop:
                    logger.warning(
                        "Multiple event loops for the GLib default context. "
                        "SIGINT may not be caught reliably")

            self._loop = weakref.ref(loop)

        def detach(self, loop):
            if self._loop:
                l = self._loop()
                if l == loop:
                    self._loop = None

        def _callback(self):
            if self._loop:
                l = self._loop()
                if l:
                    def interrupt(loop):
                        loop._interrupted = True
                        loop.stop()

                    l.call_soon_threadsafe(interrupt, l)
            return True

    @staticmethod
    def init_class():
        if not hasattr(BaseGLibEventLoop, "_default_sigint_handler"):
            BaseGLibEventLoop._default_sigint_handler = BaseGLibEventLoop.DefaultSigINTHandler()

    def __init__(self, glib_context=None, gtk=False, application=None):

        assert (glib_context is not None) + bool(gtk) + (application is not None) <= 1

        self._gtk = gtk
        self._application = application

        if gtk or self._application is not None:
            self._context = GLib.main_context_default()
        else:
            self._context = glib_context if glib_context else GLib.MainContext()


        self._readers = {}
        self._writers = {}
        self._sighandlers = {}
        self._chldhandlers = {}
        self._handlers = set()
        self._ready   = collections.deque()
        self._wakeup  = None
        self._will_dispatch = False
        self._loop_implem = None
        self._interrupted = False

        super().__init__()

        # install a default handler for SIGINT
        # in the default context
        if self._context == GLib.main_context_default():
            assert hasattr(self, "_default_sigint_handler"), "Must call BaseGLibEventLoop.init_class() first"
            self._default_sigint_handler.attach(self)

    def _dispatch(self):
        # This is the only place where callbacks are actually *called*. All
        # other places just add them to ready. Note: We run all currently
        # scheduled callbacks, but not any callbacks scheduled by callbacks run
        # this time around -- they will be run the next time (after another I/O
        # poll). Use an idiom that is threadsafe without using locks.

        self._will_dispatch = True

        ntodo = len(self._ready)
        for i in range(ntodo):
            handle = self._ready.popleft()
            if not handle._cancelled:
                handle._run()

        self._schedule_dispatch()
        self._will_dispatch = False

    def _schedule_dispatch(self):
        if not self._ready or self._wakeup is not None:
            return

        def wakeup_cb(self):
            self._dispatch()
            if self._ready:
                return True
            else:
                self._wakeup.destroy()
                self._wakeup = None
                return False

        self._wakeup = GLib.Timeout(0)
        self._wakeup.set_callback(wakeup_cb, self)
        self._wakeup.attach(self._context)


    def run_until_complete(self, future, **kw):
        """Run the event loop until a Future is done.

        Return the Future's result, or raise its exception.
        """

        def stop(f):
            self.stop()

        future = tasks.async(future, loop=self)
        future.add_done_callback(stop)
        try:
            self.run_forever(**kw)
        finally:
            future.remove_done_callback(stop)

        if not future.done():
            raise RuntimeError('Event loop stopped before Future completed.')

        return future.result()

    def run_forever(self, gtk=False, application=None):
        """Run the event loop until stop() is called."""

        assert not (gtk and application is not None)

        if self._loop_implem is not None:
            raise RuntimeError('Event loop is running.')

        if application is not None:
            assert self._context == GLib.main_context_default()
            lh = _GApplicationLoopImplem(self, application)
        elif gtk or self._gtk:
            lh = _GtkLoopImplem(self)
        elif self._application is not None:
            lh = _GApplicationLoopImplem(self, self._application)
            self._application = None
        else:
            lh = _GLibLoopImplem(self)

        # We do not run the callbacks immediately. We need to call them
        # when the Gtk loop is running, in case one callback calls .stop()
        self._schedule_dispatch()

        try:
            self._loop_implem = lh

            lh.run()

            if self._interrupted:
                # ._interrupted is set when SIGINT is caught be the default
                # signal handler implemented in this module.
                #
                # If no user-defined handler is registered, then the default
                # behaviour is just to raise KeyboardInterrupt
                #
                self._interrupted = False
                raise KeyboardInterrupt()
        finally:
            self._loop_implem = None

    def is_running(self):
        """Return whether the event loop is currently running."""
        return self._loop_implem is not None

    def stop(self):
        """Stop the event loop as soon as reasonable.

        Exactly how soon that is may depend on the implementation, but
        no more I/O callbacks should be scheduled.
        """
        lh = self._loop_implem
        if lh is not None:
            lh.stop()

    def close(self):
        for fd in list(self._readers):
            self.remove_reader(fd)

        for fd in list(self._writers):
            self.remove_writer(fd)

        for sig in list(self._sighandlers):
            self.remove_signal_handler(sig)

        for pid in list(self._chldhandlers):
            self._remove_child_handler(pid)

        for s in list(self._handlers):
            s.cancel()

        self._ready.clear() 

        self._default_sigint_handler.detach(self)

        super().close()

    # Methods scheduling callbacks.  All these return Handles.
    def call_soon(self, callback, *args):
        h = events.Handle(callback, args, self)
        self._ready.append(h)
        if not self._will_dispatch:
            self._schedule_dispatch()
        return h

    def call_later(self, delay, callback, *args):

        if delay <= 0:
            return self.call_soon(callback, *args)
        else:
            return GLibHandle(
                self,
                GLib.Timeout(delay*1000 if delay > 0 else 0),
                False,
                callback, args)

    def call_at(self, when, callback, *args):
        return self.call_later(when - self.time(), callback, *args)

    def time(self):
        return GLib.get_monotonic_time() / 1000000

    # Methods for interacting with threads.

#	def call_soon_threadsafe(self, callback, *args):
#		raise NotImplementedError
#
#	def run_in_executor(self, executor, callback, *args):
#		raise NotImplementedError
#
#	def set_default_executor(self, executor):
#		raise NotImplementedError

    # Network I/O methods returning Futures.

#	def getaddrinfo(self, host, port, *, family=0, type=0, proto=0, flags=0):
#		raise NotImplementedError
#
#	def getnameinfo(self, sockaddr, flags=0):
#		raise NotImplementedError
#
#	def create_connection(self, protocol_factory, host=None, port=None, *,
#						  ssl=None, family=0, proto=0, flags=0, sock=None,
#						  local_addr=None):
#		raise NotImplementedError
#
#	def start_serving(self, protocol_factory, host=None, port=None, *,
#					  family=socket.AF_UNSPEC, flags=socket.AI_PASSIVE,
#					  sock=None, backlog=100, ssl=None, reuse_address=None):
#		"""A coroutine which creates a TCP server bound to host and
#		port and whose result will be a list of socket objects which
#		will later be handled by protocol_factory.
#
#		If host is an empty string or None all interfaces are assumed
#		and a list of multiple sockets will be returned (most likely
#		one for IPv4 and another one for IPv6).
#
#		family can be set to either AF_INET or AF_INET6 to force the
#		socket to use IPv4 or IPv6. If not set it will be determined
#		from host (defaults to AF_UNSPEC).
#
#		flags is a bitmask for getaddrinfo().
#
#		sock can optionally be specified in order to use a preexisting
#		socket object.
#
#		backlog is the maximum number of queued connections passed to
#		listen() (defaults to 100).
#
#		ssl can be set to an SSLContext to enable SSL over the
#		accepted connections.
#
#		reuse_address tells the kernel to reuse a local socket in
#		TIME_WAIT state, without waiting for its natural timeout to
#		expire. If not specified will automatically be set to True on
#		UNIX.
#		"""
#		raise NotImplementedError
#
#	def stop_serving(self, sock):
#		"""Stop listening for incoming connections. Close socket."""
#		raise NotImplementedError
#
#	def create_datagram_endpoint(self, protocol_factory,
#								 local_addr=None, remote_addr=None, *,
#								 family=0, proto=0, flags=0):
#		raise NotImplementedError
#
#	def connect_read_pipe(self, protocol_factory, pipe):
#		"""Register read pipe in eventloop.
#
#		protocol_factory should instantiate object with Protocol interface.
#		pipe is file-like object already switched to nonblocking.
#		Return pair (transport, protocol), where transport support
#		ReadTransport ABC"""
#		# The reason to accept file-like object instead of just file descriptor
#		# is: we need to own pipe and close it at transport finishing
#		# Can got complicated errors if pass f.fileno(),
#		# close fd in pipe transport then close f and vise versa.
#		raise NotImplementedError
#
#	def connect_write_pipe(self, protocol_factory, pipe):
#		"""Register write pipe in eventloop.
#
#		protocol_factory should instantiate object with BaseProtocol interface.
#		Pipe is file-like object already switched to nonblocking.
#		Return pair (transport, protocol), where transport support
#		WriteTransport ABC"""
#		# The reason to accept file-like object instead of just file descriptor
#		# is: we need to own pipe and close it at transport finishing
#		# Can got complicated errors if pass f.fileno(),
#		# close fd in pipe transport then close f and vise versa.
#		raise NotImplementedError
#
#	def subprocess_shell(self, protocol_factory, cmd, *, stdin=subprocess.PIPE,
#						 stdout=subprocess.PIPE, stderr=subprocess.PIPE,
#						 **kwargs):
#		raise NotImplementedError
#
#	def subprocess_exec(self, protocol_factory, *args, stdin=subprocess.PIPE,
#						stdout=subprocess.PIPE, stderr=subprocess.PIPE,
#						**kwargs):
#		raise NotImplementedError

    # Ready-based callback registration methods.
    # The add_*() methods return None.
    # The remove_*() methods return True if something was removed,
    # False if there was nothing to delete.

    # FIXME: these functions are not available on windows
    def add_reader(self, fd, callback, *args):
        if not isinstance(fd, int):
            fd = fd.fileno()

        self.remove_reader(fd)

        s = GLib.unix_fd_source_new(fd, GLib.IO_IN)

        assert fd not in self._readers
        self._readers[fd] = GLibHandle(self, s, True, callback, args)

    def remove_reader(self, fd):
        if not isinstance(fd, int):
            fd = fd.fileno()

        try:
            self._readers.pop(fd).cancel()
            return True

        except KeyError:
            return False

    def add_writer(self, fd, callback, *args):
        if not isinstance(fd, int):
            fd = fd.fileno()

        self.remove_writer(fd)

        s = GLib.unix_fd_source_new(fd, GLib.IO_OUT)

        assert fd not in self._writers
        self._writers[fd] = GLibHandle(self, s, True, callback, args)

    def remove_writer(self, fd):
        if not isinstance(fd, int):
            fd = fd.fileno()

        try:
            self._writers.pop(fd).cancel()
            return True

        except KeyError:
            return False

    # Completion based I/O methods returning Futures.

#	def sock_recv(self, sock, nbytes):
#		raise NotImplementedError
#
#	def sock_sendall(self, sock, data):
#		raise NotImplementedError
#
#	def sock_connect(self, sock, address):
#		raise NotImplementedError
#
#	def sock_accept(self, sock):
#		raise NotImplementedError

    # Signal handling.

    def add_signal_handler(self, sig, callback, *args):
        self._check_signal(sig)
        self.remove_signal_handler(sig)

        s = GLib.unix_signal_source_new(sig)
        if s is None:
            if sig == signal.SIGKILL:
                raise RuntimeError("cannot catch SIGKILL")
            else:
                raise ValueError("signal not supported")

        assert sig not in self._sighandlers
        self._sighandlers[sig] = GLibHandle(self, s, True, callback, args)

    def remove_signal_handler(self, sig):
        self._check_signal(sig)
        try:
            self._sighandlers.pop(sig).cancel()
            return True

        except KeyError:
            return False

#TODO: move it into unix_events
GLibEventLoop = BaseGLibEventLoop


class _LoopImplem:
    def __init__(self, loop):
        self._loop = loop

    def run(self):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()

class _GLibLoopImplem(_LoopImplem):

    def __init__(self, loop):
        super().__init__(loop)

        # We use the introspected MainLoop object directly, because the
        # override in pygobject tampers with SIGINT
        self._mainloop = GLib._introspection_module.MainLoop.new(self._loop._context, True)

    def run(self):
        self._mainloop.run()

    def stop(self):
        self._mainloop.quit()

if Gtk:
    class _GtkLoopImplem(_LoopImplem):
        def run(self):
            Gtk.main()

        def stop(self):
            Gtk.main_quit()

class _GApplicationLoopImplem(_LoopImplem):

    def __init__(self, loop, application):
        super().__init__(loop)

        if not isinstance (application, Gio.Application):
            raise TypeError("application must be a Gio.Application object")

        self._application = application

    def run(self):
        self._application.run(None)

    def stop(self):
        self._application.quit()

class GLibEventLoopPolicy(events.AbstractEventLoopPolicy):
    """Default GLib event loop policy

    In this policy, each thread has its own event loop.  However, we only
    automatically create an event loop by default for the main thread; other
    threads by default have no event loop.
    """

    #TODO add a parameter to synchronise with GLib's thead default contextes
    #	(g_main_context_push_thread_default())
    def __init__(self, *, full=False, default=True, threads=True):
        """Constructor

        threads     Multithread support (default: True)

            Indicates whether you indend to use multiple python threads in your
            application. When set this flags disables some optimisations
            related to the GIL. (see GObject.threads_init())

        full        Full GLib (default: False)

            By default the policy is to create a GLibEventLoop object only for
            the main thread. Other threads will use regular asyncio event loops.
            If this flag is set, then this policy will use a glib event loop
            for every thread. Use this parameter if you want your loops to
            interact with modules written in other languages.

        default     Use the default context (default: True)

            Indicates whether you want to use the GLib default context. If set,
            then the loop associated with the main thread will use the default
            (NULL) GLib context (instead of creating a new one).
        """
        self._full    = full
        self._default = default

        self._default_loop = None

        self._policy  = unix_events.DefaultEventLoopPolicy()
        self._policy.new_event_loop = self.new_event_loop

        self.get_event_loop = self._policy.get_event_loop
        self.set_event_loop = self._policy.set_event_loop
        self.get_child_watcher = self._policy.get_child_watcher

        self._policy.set_child_watcher(GLibChildWatcher())

        BaseGLibEventLoop.init_class()

        if threads:
            logger.info("GLib threads enabled")
            GObject.threads_init()
        else:
            logger.info("GLib threads not used")

            def __new__(cls, *k, **kw):
                raise RuntimeError("GLib threads not enabled (you should use %s(threads=True)" % self.__class__.__name__)

            threading.Thread.__new__ = __new__

    def new_event_loop(self):
        if self._default and isinstance(threading.current_thread(), threading._MainThread):
            l = self.get_default_loop()
        elif self._full:
            l = GLibEventLoop()
        else:
            l = unix_events.DefaultEventLoopPolicy.new_event_loop(self._policy)

        return l

    def get_default_loop(self):

        if not self._default_loop:
            if not self._default:
                raise RuntimeError("%s configured not to used a default loop" % self.__class__.__name__)

            self._default_loop = self._new_default_loop()

        return self._default_loop

    def _new_default_loop(self):
        return GLibEventLoop(GLib.main_context_default())

if Gtk:
    class GtkEventLoopPolicy(GLibEventLoopPolicy):
        def __init__(self, *, full=False, threads=True):
            super().__init__ (default=True, full=full, threads=threads)

        def _new_default_loop(self):
            return GLibEventLoop(gtk=True)

class GApplicationEventLoopPolicy(GLibEventLoopPolicy):
    def __init__(self, *, full=False, threads=True):
        super().__init__ (default=True, full=full, threads=threads)

class wait_signal (futures.Future):
    def __init__(self, obj, name, *, loop=None):
        super().__init__(loop=loop)
        self._obj = obj
        #FIXME: use  a weakref ?
        self._hnd = obj.connect(name, self._signal_callback)
    
    def _signal_callback(self, *k):
        self._obj.disconnect(self._hnd)
        self.set_result(k)

    def cancel(self):
        super().cancel()
        if self._obj:
            self._obj.disconnect(self._hnd)
            self._obj = None

def get_default_loop(self):
    return asyncio.get_event_loop_policy().get_default_loop()




"""Selector eventloop for Unix with signal handling."""

import errno
import fcntl
import os
import signal
import socket
import stat
import subprocess
import sys
import threading


from asyncio import base_subprocess
from asyncio import constants
from asyncio import events
from asyncio import protocols
from .     import selector_events
from asyncio import tasks
from asyncio import transports
from asyncio.log import logger

from asyncio.unix_events import AbstractChildWatcher, DefaultEventLoopPolicy

__all__ = ['SelectorEventLoop',
           'AbstractChildWatcher', 'SafeChildWatcher',
           'FastChildWatcher', 'DefaultEventLoopPolicy',
           ]

STDIN = 0
STDOUT = 1
STDERR = 2


#if sys.platform == 'win32':  # pragma: no cover
#    raise ImportError('Signals are not really supported on Windows')


class _UnixSelectorEventLoop(selector_events.BaseSelectorEventLoop):
    """Unix event loop

    Adds signal handling to SelectorEventLoop
    """

#    def __init__(self):
#        super().__init__()
#        self._signal_handlers = {}

    def _socketpair(self):
        return socket.socketpair()

#    def close(self):
#        for sig in list(self._signal_handlers):
#            self.remove_signal_handler(sig)
#        super().close()
#
#    def add_signal_handler(self, sig, callback, *args):
#        """Add a handler for a signal.  UNIX only.
#
#        Raise ValueError if the signal number is invalid or uncatchable.
#        Raise RuntimeError if there is a problem setting up the handler.
#        """
#        self._check_signal(sig)
#        try:
#            # set_wakeup_fd() raises ValueError if this is not the
#            # main thread.  By calling it early we ensure that an
#            # event loop running in another thread cannot add a signal
#            # handler.
#            signal.set_wakeup_fd(self._csock.fileno())
#        except ValueError as exc:
#            raise RuntimeError(str(exc))
#
#        handle = events.Handle(callback, args)
#        self._signal_handlers[sig] = handle
#
#        try:
#            signal.signal(sig, self._handle_signal)
#            # Set SA_RESTART to limit EINTR occurrences.
#            signal.siginterrupt(sig, False)
#        except OSError as exc:
#            del self._signal_handlers[sig]
#            if not self._signal_handlers:
#                try:
#                    signal.set_wakeup_fd(-1)
#                except ValueError as nexc:
#                    logger.info('set_wakeup_fd(-1) failed: %s', nexc)
#
#            if exc.errno == errno.EINVAL:
#                raise RuntimeError('sig {} cannot be caught'.format(sig))
#            else:
#                raise
#
#    def _handle_signal(self, sig, arg):
#        """Internal helper that is the actual signal handler."""
#        handle = self._signal_handlers.get(sig)
#        if handle is None:
#            return  # Assume it's some race condition.
#        if handle._cancelled:
#            self.remove_signal_handler(sig)  # Remove it properly.
#        else:
#            self._add_callback_signalsafe(handle)
#
#    def remove_signal_handler(self, sig):
#        """Remove a handler for a signal.  UNIX only.
#
#        Return True if a signal handler was removed, False if not.
#        """
#        self._check_signal(sig)
#        try:
#            del self._signal_handlers[sig]
#        except KeyError:
#            return False
#
#        if sig == signal.SIGINT:
#            handler = signal.default_int_handler
#        else:
#            handler = signal.SIG_DFL
#
#        try:
#            signal.signal(sig, handler)
#        except OSError as exc:
#            if exc.errno == errno.EINVAL:
#                raise RuntimeError('sig {} cannot be caught'.format(sig))
#            else:
#                raise
#
#        if not self._signal_handlers:
#            try:
#                signal.set_wakeup_fd(-1)
#            except ValueError as exc:
#                logger.info('set_wakeup_fd(-1) failed: %s', exc)
#
#        return True

    def _check_signal(self, sig):
        """Internal helper to validate a signal.

        Raise ValueError if the signal number is invalid or uncatchable.
        Raise RuntimeError if there is a problem setting up the handler.
        """
        if not isinstance(sig, int):
            raise TypeError('sig must be an int, not {!r}'.format(sig))

        if not (1 <= sig < signal.NSIG):
            raise ValueError(
                'sig {} out of range(1, {})'.format(sig, signal.NSIG))

    def _make_read_pipe_transport(self, pipe, protocol, waiter=None,
                                  extra=None):
        return _UnixReadPipeTransport(self, pipe, protocol, waiter, extra)

    def _make_write_pipe_transport(self, pipe, protocol, waiter=None,
                                   extra=None):
        return _UnixWritePipeTransport(self, pipe, protocol, waiter, extra)

    @tasks.coroutine
    def _make_subprocess_transport(self, protocol, args, shell,
                                   stdin, stdout, stderr, bufsize,
                                   extra=None, **kwargs):
        with events.get_child_watcher() as watcher:
            transp = _UnixSubprocessTransport(self, protocol, args, shell,
                                              stdin, stdout, stderr, bufsize,
                                              extra=extra, **kwargs)
            yield from transp._post_init()
            watcher.add_child_handler(transp.get_pid(),
                                      self._child_watcher_callback, transp)

        return transp

    def _child_watcher_callback(self, pid, returncode, transp):
        self.call_soon_threadsafe(transp._process_exited, returncode)


def _set_nonblocking(fd):
    flags = fcntl.fcntl(fd, fcntl.F_GETFL)
    flags = flags | os.O_NONBLOCK
    fcntl.fcntl(fd, fcntl.F_SETFL, flags)


class _UnixReadPipeTransport(transports.ReadTransport):

    max_size = 256 * 1024  # max bytes we read in one eventloop iteration

    def __init__(self, loop, pipe, protocol, waiter=None, extra=None):
        super().__init__(extra)
        self._extra['pipe'] = pipe
        self._loop = loop
        self._pipe = pipe
        self._fileno = pipe.fileno()
        mode = os.fstat(self._fileno).st_mode
        if not (stat.S_ISFIFO(mode) or
                stat.S_ISSOCK(mode) or
                stat.S_ISCHR(mode)):
            raise ValueError("Pipe transport is for pipes/sockets only.")
        _set_nonblocking(self._fileno)
        self._protocol = protocol
        self._closing = False
        self._loop.add_reader(self._fileno, self._read_ready)
        self._loop.call_soon(self._protocol.connection_made, self)
        if waiter is not None:
            self._loop.call_soon(waiter.set_result, None)

    def _read_ready(self):
        try:
            data = os.read(self._fileno, self.max_size)
        except (BlockingIOError, InterruptedError):
            pass
        except OSError as exc:
            self._fatal_error(exc)
        else:
            if data:
                self._protocol.data_received(data)
            else:
                self._closing = True
                self._loop.remove_reader(self._fileno)
                self._loop.call_soon(self._protocol.eof_received)
                self._loop.call_soon(self._call_connection_lost, None)

    def pause_reading(self):
        self._loop.remove_reader(self._fileno)

    def resume_reading(self):
        self._loop.add_reader(self._fileno, self._read_ready)

    def close(self):
        if not self._closing:
            self._close(None)

    def _fatal_error(self, exc):
        # should be called by exception handler only
        if not (isinstance(exc, OSError) and exc.errno == errno.EIO):
            logger.exception('Fatal error for %s', self)
        self._close(exc)

    def _close(self, exc):
        self._closing = True
        self._loop.remove_reader(self._fileno)
        self._loop.call_soon(self._call_connection_lost, exc)

    def _call_connection_lost(self, exc):
        try:
            self._protocol.connection_lost(exc)
        finally:
            self._pipe.close()
            self._pipe = None
            self._protocol = None
            self._loop = None


class _UnixWritePipeTransport(selector_events._FlowControlMixin,
                              transports.WriteTransport):

    def __init__(self, loop, pipe, protocol, waiter=None, extra=None):
        super().__init__(extra)
        self._extra['pipe'] = pipe
        self._loop = loop
        self._pipe = pipe
        self._fileno = pipe.fileno()
        mode = os.fstat(self._fileno).st_mode
        is_socket = stat.S_ISSOCK(mode)
        if not (is_socket or
                stat.S_ISFIFO(mode) or
                stat.S_ISCHR(mode)):
            raise ValueError("Pipe transport is only for "
                             "pipes, sockets and character devices")
        _set_nonblocking(self._fileno)
        self._protocol = protocol
        self._buffer = []
        self._conn_lost = 0
        self._closing = False  # Set when close() or write_eof() called.

        # On AIX, the reader trick only works for sockets.
        # On other platforms it works for pipes and sockets.
        # (Exception: OS X 10.4?  Issue #19294.)
        if is_socket or not sys.platform.startswith("aix"):
            self._loop.add_reader(self._fileno, self._read_ready)

        self._loop.call_soon(self._protocol.connection_made, self)
        if waiter is not None:
            self._loop.call_soon(waiter.set_result, None)

    def get_write_buffer_size(self):
        return sum(len(data) for data in self._buffer)

    def _read_ready(self):
        # Pipe was closed by peer.
        if self._buffer:
            self._close(BrokenPipeError())
        else:
            self._close()

    def write(self, data):
        assert isinstance(data, (bytes, bytearray, memoryview)), repr(data)
        if isinstance(data, bytearray):
            data = memoryview(data)
        if not data:
            return

        if self._conn_lost or self._closing:
            if self._conn_lost >= constants.LOG_THRESHOLD_FOR_CONNLOST_WRITES:
                logger.warning('pipe closed by peer or '
                               'os.write(pipe, data) raised exception.')
            self._conn_lost += 1
            return

        if not self._buffer:
            # Attempt to send it right away first.
            try:
                n = os.write(self._fileno, data)
            except (BlockingIOError, InterruptedError):
                n = 0
            except Exception as exc:
                self._conn_lost += 1
                self._fatal_error(exc)
                return
            if n == len(data):
                return
            elif n > 0:
                data = data[n:]
            self._loop.add_writer(self._fileno, self._write_ready)

        self._buffer.append(data)
        self._maybe_pause_protocol()

    def _write_ready(self):
        data = b''.join(self._buffer)
        assert data, 'Data should not be empty'

        self._buffer.clear()
        try:
            n = os.write(self._fileno, data)
        except (BlockingIOError, InterruptedError):
            self._buffer.append(data)
        except Exception as exc:
            self._conn_lost += 1
            # Remove writer here, _fatal_error() doesn't it
            # because _buffer is empty.
            self._loop.remove_writer(self._fileno)
            self._fatal_error(exc)
        else:
            if n == len(data):
                self._loop.remove_writer(self._fileno)
                self._maybe_resume_protocol()  # May append to buffer.
                if not self._buffer and self._closing:
                    self._loop.remove_reader(self._fileno)
                    self._call_connection_lost(None)
                return
            elif n > 0:
                data = data[n:]

            self._buffer.append(data)  # Try again later.

    def can_write_eof(self):
        return True

    # TODO: Make the relationships between write_eof(), close(),
    # abort(), _fatal_error() and _close() more straightforward.

    def write_eof(self):
        if self._closing:
            return
        assert self._pipe
        self._closing = True
        if not self._buffer:
            self._loop.remove_reader(self._fileno)
            self._loop.call_soon(self._call_connection_lost, None)

    def close(self):
        if not self._closing:
            # write_eof is all what we needed to close the write pipe
            self.write_eof()

    def abort(self):
        self._close(None)

    def _fatal_error(self, exc):
        # should be called by exception handler only
        if not isinstance(exc, (BrokenPipeError, ConnectionResetError)):
            logger.exception('Fatal error for %s', self)
        self._close(exc)

    def _close(self, exc=None):
        self._closing = True
        if self._buffer:
            self._loop.remove_writer(self._fileno)
        self._buffer.clear()
        self._loop.remove_reader(self._fileno)
        self._loop.call_soon(self._call_connection_lost, exc)

    def _call_connection_lost(self, exc):
        try:
            self._protocol.connection_lost(exc)
        finally:
            self._pipe.close()
            self._pipe = None
            self._protocol = None
            self._loop = None


class _UnixSubprocessTransport(base_subprocess.BaseSubprocessTransport):

    def _start(self, args, shell, stdin, stdout, stderr, bufsize, **kwargs):
        stdin_w = None
        if stdin == subprocess.PIPE:
            # Use a socket pair for stdin, since not all platforms
            # support selecting read events on the write end of a
            # socket (which we use in order to detect closing of the
            # other end).  Notably this is needed on AIX, and works
            # just fine on other platforms.
            stdin, stdin_w = self._loop._socketpair()
        self._proc = subprocess.Popen(
            args, shell=shell, stdin=stdin, stdout=stdout, stderr=stderr,
            universal_newlines=False, bufsize=bufsize, **kwargs)
        if stdin_w is not None:
            stdin.close()
            self._proc.stdin = open(stdin_w.detach(), 'rb', buffering=bufsize)


#class AbstractChildWatcher:
#    """Abstract base class for monitoring child processes.
#
#    Objects derived from this class monitor a collection of subprocesses and
#    report their termination or interruption by a signal.
#
#    New callbacks are registered with .add_child_handler(). Starting a new
#    process must be done within a 'with' block to allow the watcher to suspend
#    its activity until the new process if fully registered (this is needed to
#    prevent a race condition in some implementations).
#
#    Example:
#        with watcher:
#            proc = subprocess.Popen("sleep 1")
#            watcher.add_child_handler(proc.pid, callback)
#
#    Notes:
#        Implementations of this class must be thread-safe.
#
#        Since child watcher objects may catch the SIGCHLD signal and call
#        waitpid(-1), there should be only one active object per process.
#    """
#
#    def add_child_handler(self, pid, callback, *args):
#        """Register a new child handler.
#
#        Arrange for callback(pid, returncode, *args) to be called when
#        process 'pid' terminates. Specifying another callback for the same
#        process replaces the previous handler.
#
#        Note: callback() must be thread-safe
#        """
#        raise NotImplementedError()
#
#    def remove_child_handler(self, pid):
#        """Removes the handler for process 'pid'.
#
#        The function returns True if the handler was successfully removed,
#        False if there was nothing to remove."""
#
#        raise NotImplementedError()
#
#    def attach_loop(self, loop):
#        """Attach the watcher to an event loop.
#
#        If the watcher was previously attached to an event loop, then it is
#        first detached before attaching to the new loop.
#
#        Note: loop may be None.
#        """
#        raise NotImplementedError()
#
#    def close(self):
#        """Close the watcher.
#
#        This must be called to make sure that any underlying resource is freed.
#        """
#        raise NotImplementedError()
#
#    def __enter__(self):
#        """Enter the watcher's context and allow starting new processes
#
#        This function must return self"""
#        raise NotImplementedError()
#
#    def __exit__(self, a, b, c):
#        """Exit the watcher's context"""
#        raise NotImplementedError()
#
#
#class BaseChildWatcher(AbstractChildWatcher):
#
#    def __init__(self):
#        self._loop = None
#
#    def close(self):
#        self.attach_loop(None)
#
#    def _do_waitpid(self, expected_pid):
#        raise NotImplementedError()
#
#    def _do_waitpid_all(self):
#        raise NotImplementedError()
#
#    def attach_loop(self, loop):
#        assert loop is None or isinstance(loop, events.AbstractEventLoop)
#
#        if self._loop is not None:
#            self._loop.remove_signal_handler(signal.SIGCHLD)
#
#        self._loop = loop
#        if loop is not None:
#            loop.add_signal_handler(signal.SIGCHLD, self._sig_chld)
#
#            # Prevent a race condition in case a child terminated
#            # during the switch.
#            self._do_waitpid_all()
#
#    def _sig_chld(self):
#        try:
#            self._do_waitpid_all()
#        except Exception:
#            logger.exception('Unknown exception in SIGCHLD handler')
#
#    def _compute_returncode(self, status):
#        if os.WIFSIGNALED(status):
#            # The child process died because of a signal.
#            return -os.WTERMSIG(status)
#        elif os.WIFEXITED(status):
#            # The child process exited (e.g sys.exit()).
#            return os.WEXITSTATUS(status)
#        else:
#            # The child exited, but we don't understand its status.
#            # This shouldn't happen, but if it does, let's just
#            # return that status; perhaps that helps debug it.
#            return status
#
#
#class SafeChildWatcher(BaseChildWatcher):
#    """'Safe' child watcher implementation.
#
#    This implementation avoids disrupting other code spawning processes by
#    polling explicitly each process in the SIGCHLD handler instead of calling
#    os.waitpid(-1).
#
#    This is a safe solution but it has a significant overhead when handling a
#    big number of children (O(n) each time SIGCHLD is raised)
#    """
#
#    def __init__(self):
#        super().__init__()
#        self._callbacks = {}
#
#    def close(self):
#        self._callbacks.clear()
#        super().close()
#
#    def __enter__(self):
#        return self
#
#    def __exit__(self, a, b, c):
#        pass
#
#    def add_child_handler(self, pid, callback, *args):
#        self._callbacks[pid] = callback, args
#
#        # Prevent a race condition in case the child is already terminated.
#        self._do_waitpid(pid)
#
#    def remove_child_handler(self, pid):
#        try:
#            del self._callbacks[pid]
#            return True
#        except KeyError:
#            return False
#
#    def _do_waitpid_all(self):
#
#        for pid in list(self._callbacks):
#            self._do_waitpid(pid)
#
#    def _do_waitpid(self, expected_pid):
#        assert expected_pid > 0
#
#        try:
#            pid, status = os.waitpid(expected_pid, os.WNOHANG)
#        except ChildProcessError:
#            # The child process is already reaped
#            # (may happen if waitpid() is called elsewhere).
#            pid = expected_pid
#            returncode = 255
#            logger.warning(
#                "Unknown child process pid %d, will report returncode 255",
#                pid)
#        else:
#            if pid == 0:
#                # The child process is still alive.
#                return
#
#            returncode = self._compute_returncode(status)
#
#        try:
#            callback, args = self._callbacks.pop(pid)
#        except KeyError:  # pragma: no cover
#            # May happen if .remove_child_handler() is called
#            # after os.waitpid() returns.
#            pass
#        else:
#            callback(pid, returncode, *args)
#
#
#class FastChildWatcher(BaseChildWatcher):
#    """'Fast' child watcher implementation.
#
#    This implementation reaps every terminated processes by calling
#    os.waitpid(-1) directly, possibly breaking other code spawning processes
#    and waiting for their termination.
#
#    There is no noticeable overhead when handling a big number of children
#    (O(1) each time a child terminates).
#    """
#    def __init__(self):
#        super().__init__()
#        self._callbacks = {}
#        self._lock = threading.Lock()
#        self._zombies = {}
#        self._forks = 0
#
#    def close(self):
#        self._callbacks.clear()
#        self._zombies.clear()
#        super().close()
#
#    def __enter__(self):
#        with self._lock:
#            self._forks += 1
#
#            return self
#
#    def __exit__(self, a, b, c):
#        with self._lock:
#            self._forks -= 1
#
#            if self._forks or not self._zombies:
#                return
#
#            collateral_victims = str(self._zombies)
#            self._zombies.clear()
#
#        logger.warning(
#            "Caught subprocesses termination from unknown pids: %s",
#            collateral_victims)
#
#    def add_child_handler(self, pid, callback, *args):
#        assert self._forks, "Must use the context manager"
#        with self._lock:
#            try:
#                returncode = self._zombies.pop(pid)
#            except KeyError:
#                # The child is running.
#                self._callbacks[pid] = callback, args
#                return
#
#        # The child is dead already. We can fire the callback.
#        callback(pid, returncode, *args)
#
#    def remove_child_handler(self, pid):
#        try:
#            del self._callbacks[pid]
#            return True
#        except KeyError:
#            return False
#
#    def _do_waitpid_all(self):
#        # Because of signal coalescing, we must keep calling waitpid() as
#        # long as we're able to reap a child.
#        while True:
#            try:
#                pid, status = os.waitpid(-1, os.WNOHANG)
#            except ChildProcessError:
#                # No more child processes exist.
#                return
#            else:
#                if pid == 0:
#                    # A child process is still alive.
#                    return
#
#                returncode = self._compute_returncode(status)
#
#            with self._lock:
#                try:
#                    callback, args = self._callbacks.pop(pid)
#                except KeyError:
#                    # unknown child
#                    if self._forks:
#                        # It may not be registered yet.
#                        self._zombies[pid] = returncode
#                        continue
#                    callback = None
#
#            if callback is None:
#                logger.warning(
#                    "Caught subprocess termination from unknown pid: "
#                    "%d -> %d", pid, returncode)
#            else:
#                callback(pid, returncode, *args)
#
#
#class _UnixDefaultEventLoopPolicy(events.BaseDefaultEventLoopPolicy):
#    """XXX"""
#    _loop_factory = _UnixSelectorEventLoop
#
#    def __init__(self):
#        super().__init__()
#        self._watcher = None
#
#    def _init_watcher(self):
#        with events._lock:
#            if self._watcher is None:  # pragma: no branch
#                self._watcher = SafeChildWatcher()
#                if isinstance(threading.current_thread(),
#                              threading._MainThread):
#                    self._watcher.attach_loop(self._local._loop)
#
#    def set_event_loop(self, loop):
#        """Set the event loop.
#
#        As a side effect, if a child watcher was set before, then calling
#        .set_event_loop() from the main thread will call .attach_loop(loop) on
#        the child watcher.
#        """
#
#        super().set_event_loop(loop)
#
#        if self._watcher is not None and \
#            isinstance(threading.current_thread(), threading._MainThread):
#            self._watcher.attach_loop(loop)
#
#    def get_child_watcher(self):
#        """Get the child watcher
#
#        If not yet set, a SafeChildWatcher object is automatically created.
#        """
#        if self._watcher is None:
#            self._init_watcher()
#
#        return self._watcher
#
#    def set_child_watcher(self, watcher):
#        """Set the child watcher"""
#
#        assert watcher is None or isinstance(watcher, AbstractChildWatcher)
#
#        if self._watcher is not None:
#            self._watcher.close()
#
#        self._watcher = watcher

SelectorEventLoop = _UnixSelectorEventLoop
#DefaultEventLoopPolicy = _UnixDefaultEventLoopPolicy



"""Event loop using a selector and related classes.

A selector is a "notify-when-ready" multiplexer.  For a subclass which
also includes support for signal handling, see the unix_events sub-module.
"""

__all__ = ['BaseSelectorEventLoop']

import collections
import errno
import socket
try:
    import ssl
except ImportError:  # pragma: no cover
    ssl = None

from . import base_events
from asyncio import constants
from asyncio import events
from asyncio import futures
from asyncio import selectors
from asyncio import transports
from asyncio.log import logger


class BaseSelectorEventLoop(base_events.BaseEventLoop):
    """Selector event loop.

    See events.EventLoop for API specification.
    """

    def __init__(self, selector=None):
        super().__init__()

#        if selector is None:
#            selector = selectors.DefaultSelector()
#        logger.debug('Using selector: %s', selector.__class__.__name__)
#        self._selector = selector
        self._make_self_pipe()

    def _make_socket_transport(self, sock, protocol, waiter=None, *,
                               extra=None, server=None):
        return _SelectorSocketTransport(self, sock, protocol, waiter,
                                        extra, server)

    def _make_ssl_transport(self, rawsock, protocol, sslcontext, waiter, *,
                            server_side=False, server_hostname=None,
                            extra=None, server=None):
        return _SelectorSslTransport(
            self, rawsock, protocol, sslcontext, waiter,
            server_side, server_hostname, extra, server)

    def _make_datagram_transport(self, sock, protocol,
                                 address=None, extra=None):
        return _SelectorDatagramTransport(self, sock, protocol, address, extra)

    def close(self):
        if self._ssock is not None:
#        if self._selector is not None:
            self._close_self_pipe()
#            self._selector.close()
#            self._selector = None
            super().close()

    def _socketpair(self):
        raise NotImplementedError

    def _close_self_pipe(self):
        self.remove_reader(self._ssock.fileno())
        self._ssock.close()
        self._ssock = None
        self._csock.close()
        self._csock = None
        self._internal_fds -= 1

    def _make_self_pipe(self):
        # A self-socket, really. :-)
        self._ssock, self._csock = self._socketpair()
        self._ssock.setblocking(False)
        self._csock.setblocking(False)
        self._internal_fds += 1
        self.add_reader(self._ssock.fileno(), self._read_from_self)

    def _read_from_self(self):
        try:
            self._ssock.recv(1)
        except (BlockingIOError, InterruptedError):
            pass

    def _write_to_self(self):
        try:
            self._csock.send(b'x')
        except (BlockingIOError, InterruptedError):
            pass

    def _start_serving(self, protocol_factory, sock,
                       sslcontext=None, server=None):
        self.add_reader(sock.fileno(), self._accept_connection,
                        protocol_factory, sock, sslcontext, server)

    def _accept_connection(self, protocol_factory, sock,
                           sslcontext=None, server=None):
        try:
            conn, addr = sock.accept()
            conn.setblocking(False)
        except (BlockingIOError, InterruptedError, ConnectionAbortedError):
            pass  # False alarm.
        except OSError as exc:
            # There's nowhere to send the error, so just log it.
            # TODO: Someone will want an error handler for this.
            if exc.errno in (errno.EMFILE, errno.ENFILE,
                             errno.ENOBUFS, errno.ENOMEM):
                # Some platforms (e.g. Linux keep reporting the FD as
                # ready, so we remove the read handler temporarily.
                # We'll try again in a while.
                logger.exception('Accept out of system resource (%s)', exc)
                self.remove_reader(sock.fileno())
                self.call_later(constants.ACCEPT_RETRY_DELAY,
                                self._start_serving,
                                protocol_factory, sock, sslcontext, server)
            else:
                raise  # The event loop will catch, log and ignore it.
        else:
            if sslcontext:
                self._make_ssl_transport(
                    conn, protocol_factory(), sslcontext, None,
                    server_side=True, extra={'peername': addr}, server=server)
            else:
                self._make_socket_transport(
                    conn, protocol_factory(), extra={'peername': addr},
                    server=server)
        # It's now up to the protocol to handle the connection.

#    def add_reader(self, fd, callback, *args):
#        """Add a reader callback."""
#        handle = events.Handle(callback, args)
#        try:
#            key = self._selector.get_key(fd)
#        except KeyError:
#            self._selector.register(fd, selectors.EVENT_READ,
#                                    (handle, None))
#        else:
#            mask, (reader, writer) = key.events, key.data
#            self._selector.modify(fd, mask | selectors.EVENT_READ,
#                                  (handle, writer))
#            if reader is not None:
#                reader.cancel()
#
#    def remove_reader(self, fd):
#        """Remove a reader callback."""
#        try:
#            key = self._selector.get_key(fd)
#        except KeyError:
#            return False
#        else:
#            mask, (reader, writer) = key.events, key.data
#            mask &= ~selectors.EVENT_READ
#            if not mask:
#                self._selector.unregister(fd)
#            else:
#                self._selector.modify(fd, mask, (None, writer))
#
#            if reader is not None:
#                reader.cancel()
#                return True
#            else:
#                return False
#
#    def add_writer(self, fd, callback, *args):
#        """Add a writer callback.."""
#        handle = events.Handle(callback, args)
#        try:
#            key = self._selector.get_key(fd)
#        except KeyError:
#            self._selector.register(fd, selectors.EVENT_WRITE,
#                                    (None, handle))
#        else:
#            mask, (reader, writer) = key.events, key.data
#            self._selector.modify(fd, mask | selectors.EVENT_WRITE,
#                                  (reader, handle))
#            if writer is not None:
#                writer.cancel()
#
#    def remove_writer(self, fd):
#        """Remove a writer callback."""
#        try:
#            key = self._selector.get_key(fd)
#        except KeyError:
#            return False
#        else:
#            mask, (reader, writer) = key.events, key.data
#            # Remove both writer and connector.
#            mask &= ~selectors.EVENT_WRITE
#            if not mask:
#                self._selector.unregister(fd)
#            else:
#                self._selector.modify(fd, mask, (reader, None))
#
#            if writer is not None:
#                writer.cancel()
#                return True
#            else:
#                return False

    def sock_recv(self, sock, n):
        """XXX"""
        fut = futures.Future(loop=self)
        self._sock_recv(fut, False, sock, n)
        return fut

    def _sock_recv(self, fut, registered, sock, n):
        fd = sock.fileno()
        if registered:
            # Remove the callback early.  It should be rare that the
            # selector says the fd is ready but the call still returns
            # EAGAIN, and I am willing to take a hit in that case in
            # order to simplify the common case.
            self.remove_reader(fd)
        if fut.cancelled():
            return
        try:
            data = sock.recv(n)
        except (BlockingIOError, InterruptedError):
            self.add_reader(fd, self._sock_recv, fut, True, sock, n)
        except Exception as exc:
            fut.set_exception(exc)
        else:
            fut.set_result(data)

    def sock_sendall(self, sock, data):
        """XXX"""
        fut = futures.Future(loop=self)
        if data:
            self._sock_sendall(fut, False, sock, data)
        else:
            fut.set_result(None)
        return fut

    def _sock_sendall(self, fut, registered, sock, data):
        fd = sock.fileno()

        if registered:
            self.remove_writer(fd)
        if fut.cancelled():
            return

        try:
            n = sock.send(data)
        except (BlockingIOError, InterruptedError):
            n = 0
        except Exception as exc:
            fut.set_exception(exc)
            return

        if n == len(data):
            fut.set_result(None)
        else:
            if n:
                data = data[n:]
            self.add_writer(fd, self._sock_sendall, fut, True, sock, data)

    def sock_connect(self, sock, address):
        """XXX"""
        # That address better not require a lookup!  We're not calling
        # self.getaddrinfo() for you here.  But verifying this is
        # complicated; the socket module doesn't have a pattern for
        # IPv6 addresses (there are too many forms, apparently).
        fut = futures.Future(loop=self)
        self._sock_connect(fut, False, sock, address)
        return fut

    def _sock_connect(self, fut, registered, sock, address):
        # TODO: Use getaddrinfo() to look up the address, to avoid the
        # trap of hanging the entire event loop when the address
        # requires doing a DNS lookup.  (OTOH, the caller should
        # already have done this, so it would be nice if we could
        # easily tell whether the address needs looking up or not.  I
        # know how to do this for IPv4, but IPv6 addresses have many
        # syntaxes.)
        fd = sock.fileno()
        if registered:
            self.remove_writer(fd)
        if fut.cancelled():
            return
        try:
            if not registered:
                # First time around.
                sock.connect(address)
            else:
                err = sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
                if err != 0:
                    # Jump to the except clause below.
                    raise OSError(err, 'Connect call failed %s' % (address,))
        except (BlockingIOError, InterruptedError):
            self.add_writer(fd, self._sock_connect, fut, True, sock, address)
        except Exception as exc:
            fut.set_exception(exc)
        else:
            fut.set_result(None)

    def sock_accept(self, sock):
        """XXX"""
        fut = futures.Future(loop=self)
        self._sock_accept(fut, False, sock)
        return fut

    def _sock_accept(self, fut, registered, sock):
        fd = sock.fileno()
        if registered:
            self.remove_reader(fd)
        if fut.cancelled():
            return
        try:
            conn, address = sock.accept()
            conn.setblocking(False)
        except (BlockingIOError, InterruptedError):
            self.add_reader(fd, self._sock_accept, fut, True, sock)
        except Exception as exc:
            fut.set_exception(exc)
        else:
            fut.set_result((conn, address))

#    def _process_events(self, event_list):
#        for key, mask in event_list:
#            fileobj, (reader, writer) = key.fileobj, key.data
#            if mask & selectors.EVENT_READ and reader is not None:
#                if reader._cancelled:
#                    self.remove_reader(fileobj)
#                else:
#                    self._add_callback(reader)
#            if mask & selectors.EVENT_WRITE and writer is not None:
#                if writer._cancelled:
#                    self.remove_writer(fileobj)
#                else:
#                    self._add_callback(writer)

    def _stop_serving(self, sock):
        self.remove_reader(sock.fileno())
        sock.close()


class _FlowControlMixin(transports.Transport):
    """All the logic for (write) flow control in a mix-in base class.

    The subclass must implement get_write_buffer_size().  It must call
    _maybe_pause_protocol() whenever the write buffer size increases,
    and _maybe_resume_protocol() whenever it decreases.  It may also
    override set_write_buffer_limits() (e.g. to specify different
    defaults).

    The subclass constructor must call super().__init__(extra).  This
    will call set_write_buffer_limits().

    The user may call set_write_buffer_limits() and
    get_write_buffer_size(), and their protocol's pause_writing() and
    resume_writing() may be called.
    """

    def __init__(self, extra=None):
        super().__init__(extra)
        self._protocol_paused = False
        self.set_write_buffer_limits()

    def _maybe_pause_protocol(self):
        size = self.get_write_buffer_size()
        if size <= self._high_water:
            return
        if not self._protocol_paused:
            self._protocol_paused = True
            try:
                self._protocol.pause_writing()
            except Exception:
                logger.exception('pause_writing() failed')

    def _maybe_resume_protocol(self):
        if (self._protocol_paused and
            self.get_write_buffer_size() <= self._low_water):
            self._protocol_paused = False
            try:
                self._protocol.resume_writing()
            except Exception:
                logger.exception('resume_writing() failed')

    def set_write_buffer_limits(self, high=None, low=None):
        if high is None:
            if low is None:
                high = 64*1024
            else:
                high = 4*low
        if low is None:
            low = high // 4
        if not high >= low >= 0:
            raise ValueError('high (%r) must be >= low (%r) must be >= 0' %
                             (high, low))
        self._high_water = high
        self._low_water = low

    def get_write_buffer_size(self):
        raise NotImplementedError


class _SelectorTransport(_FlowControlMixin, transports.Transport):

    max_size = 256 * 1024  # Buffer size passed to recv().

    _buffer_factory = bytearray  # Constructs initial value for self._buffer.

    def __init__(self, loop, sock, protocol, extra, server=None):
        super().__init__(extra)
        self._extra['socket'] = sock
        self._extra['sockname'] = sock.getsockname()
        if 'peername' not in self._extra:
            try:
                self._extra['peername'] = sock.getpeername()
            except socket.error:
                self._extra['peername'] = None
        self._loop = loop
        self._sock = sock
        self._sock_fd = sock.fileno()
        self._protocol = protocol
        self._server = server
        self._buffer = self._buffer_factory()
        self._conn_lost = 0  # Set when call to connection_lost scheduled.
        self._closing = False  # Set when close() called.
        if self._server is not None:
            self._server.attach(self)

    def abort(self):
        self._force_close(None)

    def close(self):
        if self._closing:
            return
        self._closing = True
        self._loop.remove_reader(self._sock_fd)
        if not self._buffer:
            self._conn_lost += 1
            self._loop.call_soon(self._call_connection_lost, None)

    def _fatal_error(self, exc):
        # Should be called from exception handler only.
        if not isinstance(exc, (BrokenPipeError, ConnectionResetError)):
            logger.exception('Fatal error for %s', self)
        self._force_close(exc)

    def _force_close(self, exc):
        if self._conn_lost:
            return
        if self._buffer:
            self._buffer.clear()
            self._loop.remove_writer(self._sock_fd)
        if not self._closing:
            self._closing = True
            self._loop.remove_reader(self._sock_fd)
        self._conn_lost += 1
        self._loop.call_soon(self._call_connection_lost, exc)

    def _call_connection_lost(self, exc):
        try:
            self._protocol.connection_lost(exc)
        finally:
            self._sock.close()
            self._sock = None
            self._protocol = None
            self._loop = None
            server = self._server
            if server is not None:
                server.detach(self)
                self._server = None

    def get_write_buffer_size(self):
        return len(self._buffer)


class _SelectorSocketTransport(_SelectorTransport):

    def __init__(self, loop, sock, protocol, waiter=None,
                 extra=None, server=None):
        super().__init__(loop, sock, protocol, extra, server)
        self._eof = False
        self._paused = False

        self._loop.add_reader(self._sock_fd, self._read_ready)
        self._loop.call_soon(self._protocol.connection_made, self)
        if waiter is not None:
            self._loop.call_soon(waiter.set_result, None)

    def pause_reading(self):
        if self._closing:
            raise RuntimeError('Cannot pause_reading() when closing')
        if self._paused:
            raise RuntimeError('Already paused')
        self._paused = True
        self._loop.remove_reader(self._sock_fd)

    def resume_reading(self):
        if not self._paused:
            raise RuntimeError('Not paused')
        self._paused = False
        if self._closing:
            return
        self._loop.add_reader(self._sock_fd, self._read_ready)

    def _read_ready(self):
        try:
            data = self._sock.recv(self.max_size)
        except (BlockingIOError, InterruptedError):
            pass
        except Exception as exc:
            self._fatal_error(exc)
        else:
            if data:
                self._protocol.data_received(data)
            else:
                keep_open = self._protocol.eof_received()
                if keep_open:
                    # We're keeping the connection open so the
                    # protocol can write more, but we still can't
                    # receive more, so remove the reader callback.
                    self._loop.remove_reader(self._sock_fd)
                else:
                    self.close()

    def write(self, data):
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError('data argument must be byte-ish (%r)',
                            type(data))
        if self._eof:
            raise RuntimeError('Cannot call write() after write_eof()')
        if not data:
            return

        if self._conn_lost:
            if self._conn_lost >= constants.LOG_THRESHOLD_FOR_CONNLOST_WRITES:
                logger.warning('socket.send() raised exception.')
            self._conn_lost += 1
            return

        if not self._buffer:
            # Optimization: try to send now.
            try:
                n = self._sock.send(data)
            except (BlockingIOError, InterruptedError):
                pass
            except Exception as exc:
                self._fatal_error(exc)
                return
            else:
                data = data[n:]
                if not data:
                    return
            # Not all was written; register write handler.
            self._loop.add_writer(self._sock_fd, self._write_ready)

        # Add it to the buffer.
        self._buffer.extend(data)
        self._maybe_pause_protocol()

    def _write_ready(self):
        assert self._buffer, 'Data should not be empty'

        try:
            n = self._sock.send(self._buffer)
        except (BlockingIOError, InterruptedError):
            pass
        except Exception as exc:
            self._loop.remove_writer(self._sock_fd)
            self._buffer.clear()
            self._fatal_error(exc)
        else:
            if n:
                del self._buffer[:n]
            self._maybe_resume_protocol()  # May append to buffer.
            if not self._buffer:
                self._loop.remove_writer(self._sock_fd)
                if self._closing:
                    self._call_connection_lost(None)
                elif self._eof:
                    self._sock.shutdown(socket.SHUT_WR)

    def write_eof(self):
        if self._eof:
            return
        self._eof = True
        if not self._buffer:
            self._sock.shutdown(socket.SHUT_WR)

    def can_write_eof(self):
        return True


class _SelectorSslTransport(_SelectorTransport):

    _buffer_factory = bytearray

    def __init__(self, loop, rawsock, protocol, sslcontext, waiter=None,
                 server_side=False, server_hostname=None,
                 extra=None, server=None):
        if ssl is None:
            raise RuntimeError('stdlib ssl module not available')

        if server_side:
            if not sslcontext:
                raise ValueError('Server side ssl needs a valid SSLContext')
        else:
            if not sslcontext:
                # Client side may pass ssl=True to use a default
                # context; in that case the sslcontext passed is None.
                # The default is the same as used by urllib with
                # cadefault=True.
                if hasattr(ssl, '_create_stdlib_context'):
                    sslcontext = ssl._create_stdlib_context(
                        cert_reqs=ssl.CERT_REQUIRED,
                        check_hostname=bool(server_hostname))
                else:
                    # Fallback for Python 3.3.
                    sslcontext = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
                    sslcontext.options |= ssl.OP_NO_SSLv2
                    sslcontext.set_default_verify_paths()
                    sslcontext.verify_mode = ssl.CERT_REQUIRED

        wrap_kwargs = {
            'server_side': server_side,
            'do_handshake_on_connect': False,
        }
        if server_hostname and not server_side and ssl.HAS_SNI:
            wrap_kwargs['server_hostname'] = server_hostname
        sslsock = sslcontext.wrap_socket(rawsock, **wrap_kwargs)

        super().__init__(loop, sslsock, protocol, extra, server)

        self._server_hostname = server_hostname
        self._waiter = waiter
        self._rawsock = rawsock
        self._sslcontext = sslcontext
        self._paused = False

        # SSL-specific extra info.  (peercert is set later)
        self._extra.update(sslcontext=sslcontext)

        self._on_handshake()

    def _on_handshake(self):
        try:
            self._sock.do_handshake()
        except ssl.SSLWantReadError:
            self._loop.add_reader(self._sock_fd, self._on_handshake)
            return
        except ssl.SSLWantWriteError:
            self._loop.add_writer(self._sock_fd, self._on_handshake)
            return
        except Exception as exc:
            self._loop.remove_reader(self._sock_fd)
            self._loop.remove_writer(self._sock_fd)
            self._sock.close()
            if self._waiter is not None:
                self._waiter.set_exception(exc)
            return
        except BaseException as exc:
            self._loop.remove_reader(self._sock_fd)
            self._loop.remove_writer(self._sock_fd)
            self._sock.close()
            if self._waiter is not None:
                self._waiter.set_exception(exc)
            raise

        self._loop.remove_reader(self._sock_fd)
        self._loop.remove_writer(self._sock_fd)

        peercert = self._sock.getpeercert()
        if not hasattr(self._sslcontext, 'check_hostname'):
            # Verify hostname if requested, Python 3.4+ uses check_hostname
            # and checks the hostname in do_handshake()
            if (self._server_hostname and
                self._sslcontext.verify_mode != ssl.CERT_NONE):
                try:
                    ssl.match_hostname(peercert, self._server_hostname)
                except Exception as exc:
                    self._sock.close()
                    if self._waiter is not None:
                        self._waiter.set_exception(exc)
                    return

        # Add extra info that becomes available after handshake.
        self._extra.update(peercert=peercert,
                           cipher=self._sock.cipher(),
                           compression=self._sock.compression(),
                           )

        self._read_wants_write = False
        self._write_wants_read = False
        self._loop.add_reader(self._sock_fd, self._read_ready)
        self._loop.call_soon(self._protocol.connection_made, self)
        if self._waiter is not None:
            self._loop.call_soon(self._waiter.set_result, None)

    def pause_reading(self):
        # XXX This is a bit icky, given the comment at the top of
        # _read_ready().  Is it possible to evoke a deadlock?  I don't
        # know, although it doesn't look like it; write() will still
        # accept more data for the buffer and eventually the app will
        # call resume_reading() again, and things will flow again.

        if self._closing:
            raise RuntimeError('Cannot pause_reading() when closing')
        if self._paused:
            raise RuntimeError('Already paused')
        self._paused = True
        self._loop.remove_reader(self._sock_fd)

    def resume_reading(self):
        if not self._paused:
            raise ('Not paused')
        self._paused = False
        if self._closing:
            return
        self._loop.add_reader(self._sock_fd, self._read_ready)

    def _read_ready(self):
        if self._write_wants_read:
            self._write_wants_read = False
            self._write_ready()

            if self._buffer:
                self._loop.add_writer(self._sock_fd, self._write_ready)

        try:
            data = self._sock.recv(self.max_size)
        except (BlockingIOError, InterruptedError, ssl.SSLWantReadError):
            pass
        except ssl.SSLWantWriteError:
            self._read_wants_write = True
            self._loop.remove_reader(self._sock_fd)
            self._loop.add_writer(self._sock_fd, self._write_ready)
        except Exception as exc:
            self._fatal_error(exc)
        else:
            if data:
                self._protocol.data_received(data)
            else:
                try:
                    keep_open = self._protocol.eof_received()
                    if keep_open:
                        logger.warning('returning true from eof_received() '
                                       'has no effect when using ssl')
                finally:
                    self.close()

    def _write_ready(self):
        if self._read_wants_write:
            self._read_wants_write = False
            self._read_ready()

            if not (self._paused or self._closing):
                self._loop.add_reader(self._sock_fd, self._read_ready)

        if self._buffer:
            try:
                n = self._sock.send(self._buffer)
            except (BlockingIOError, InterruptedError,
                    ssl.SSLWantWriteError):
                n = 0
            except ssl.SSLWantReadError:
                n = 0
                self._loop.remove_writer(self._sock_fd)
                self._write_wants_read = True
            except Exception as exc:
                self._loop.remove_writer(self._sock_fd)
                self._buffer.clear()
                self._fatal_error(exc)
                return

            if n:
                del self._buffer[:n]

        self._maybe_resume_protocol()  # May append to buffer.

        if not self._buffer:
            self._loop.remove_writer(self._sock_fd)
            if self._closing:
                self._call_connection_lost(None)

    def write(self, data):
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError('data argument must be byte-ish (%r)',
                            type(data))
        if not data:
            return

        if self._conn_lost:
            if self._conn_lost >= constants.LOG_THRESHOLD_FOR_CONNLOST_WRITES:
                logger.warning('socket.send() raised exception.')
            self._conn_lost += 1
            return

        if not self._buffer:
            self._loop.add_writer(self._sock_fd, self._write_ready)

        # Add it to the buffer.
        self._buffer.extend(data)
        self._maybe_pause_protocol()

    def can_write_eof(self):
        return False


class _SelectorDatagramTransport(_SelectorTransport):

    _buffer_factory = collections.deque

    def __init__(self, loop, sock, protocol, address=None, extra=None):
        super().__init__(loop, sock, protocol, extra)
        self._address = address
        self._loop.add_reader(self._sock_fd, self._read_ready)
        self._loop.call_soon(self._protocol.connection_made, self)

    def get_write_buffer_size(self):
        return sum(len(data) for data, _ in self._buffer)

    def _read_ready(self):
        try:
            data, addr = self._sock.recvfrom(self.max_size)
        except (BlockingIOError, InterruptedError):
            pass
        except OSError as exc:
            self._protocol.error_received(exc)
        except Exception as exc:
            self._fatal_error(exc)
        else:
            self._protocol.datagram_received(data, addr)

    def sendto(self, data, addr=None):
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError('data argument must be byte-ish (%r)',
                            type(data))
        if not data:
            return

        if self._address and addr not in (None, self._address):
            raise ValueError('Invalid address: must be None or %s' %
                             (self._address,))

        if self._conn_lost and self._address:
            if self._conn_lost >= constants.LOG_THRESHOLD_FOR_CONNLOST_WRITES:
                logger.warning('socket.send() raised exception.')
            self._conn_lost += 1
            return

        if not self._buffer:
            # Attempt to send it right away first.
            try:
                if self._address:
                    self._sock.send(data)
                else:
                    self._sock.sendto(data, addr)
                return
            except (BlockingIOError, InterruptedError):
                self._loop.add_writer(self._sock_fd, self._sendto_ready)
            except OSError as exc:
                self._protocol.error_received(exc)
                return
            except Exception as exc:
                self._fatal_error(exc)
                return

        # Ensure that what we buffer is immutable.
        self._buffer.append((bytes(data), addr))
        self._maybe_pause_protocol()

    def _sendto_ready(self):
        while self._buffer:
            data, addr = self._buffer.popleft()
            try:
                if self._address:
                    self._sock.send(data)
                else:
                    self._sock.sendto(data, addr)
            except (BlockingIOError, InterruptedError):
                self._buffer.appendleft((data, addr))  # Try again later.
                break
            except OSError as exc:
                self._protocol.error_received(exc)
                return
            except Exception as exc:
                self._fatal_error(exc)
                return

        self._maybe_resume_protocol()  # May append to buffer.
        if not self._buffer:
            self._loop.remove_writer(self._sock_fd)
            if self._closing:
                self._call_connection_lost(None)



#!/usr/bin/env python

from distutils.core import setup

setup(name='gpotato',
      version='0.1',
      description='GLib event loop for tulip (PEP 3156)',
      author='Jamie Bliss',
      author_email='astronouth7303@gmail.com',
      license='Apache 2.0',
      url='https://github.com/astronouth7303/gpotato',
      packages=['gpotato'],
      data_files=['README.md', 'examples/test-gtk.py'],
      long_description="""Gbulb is a python library that implements a PEP 3156 interface for the GLib main event loop. It is designed to be used together with the tulip reference implementation.

This is a work in progress. The code is experimental and may break at any time.
""",
      classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ]
)

